/**
 * @file Retrokits base library for RK008
 * @copyright Retrokits 2021
 * @author Retrokits (A.J. Huitsing)
 * @version 0.1
 */

"use strict";

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
const RKTRACE_GLB			= {ena:true , tag:"          "};
const RKTRACE_APP			= {ena:true , tag:"APP       "};
const RKTRACE_RKDEV 		= {ena:true , tag:"RKDEV     "};
const RKTRACE_RK006 		= {ena:true , tag:"RK006     "};
const RKTRACE_WEBMIDI		= {ena:false, tag:"WEBMIDI   "};
const RKTRACE_WEBSERIAL		= {ena:true,  tag:"WEBSERIAL "};
const RKTRACE_MIKMIDI       = {ena:true , tag:"MIKMIDI   "};
const RKTRACE_RK008			= {ena:true , tag:"RK008     "};

var trace_last = null;

function RKTRACE(mod,str)
{
	if (!mod.ena)
	{
		return;
	}
	
	var dt;
	
	var trace_now = Date.now();
	
	if (trace_last == null)
	{
		dt = 0;
	}
	else
	{
		dt = trace_now - trace_last;
		if (dt>99999)
		{
			dt=99999;
		}
	}
	trace_last = trace_now;
	
	dt = dt + "      ";
	dt = dt.substring(0,6);

	console.log(dt + mod.tag + str);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
(function(factory, global) {
if (typeof define === 'function' && define.amd) {
define(function() {
return factory(global, navigator)
})
} else if (typeof exports === 'object' && typeof module === 'object') {
module.exports = factory(global, navigator)
} else {
// mock the navigator object when under test since `navigator.onLine` is read only
global.RobustWebSocket = factory(global, typeof Mocha !== 'undefined' ? Mocha : navigator)
}
})(function(global, navigator) {

var RobustWebSocket = function(url, protocols, userOptions) {
var realWs = { close: function() {} },
connectTimeout,
self = this,
attempts = 0,
reconnects = -1,
reconnectWhenOnlineAgain = false,
explicitlyClosed = false,
pendingReconnect,
opts = Object.assign({},
RobustWebSocket.defaultOptions,
typeof userOptions === 'function' ? { shouldReconnect: userOptions } : userOptions
)

if (typeof opts.timeout !== 'number') {
throw new Error('timeout must be the number of milliseconds to timeout a connection attempt')
}

if (typeof opts.shouldReconnect !== 'function') {
throw new Error('shouldReconnect must be a function that returns the number of milliseconds to wait for a reconnect attempt, or null or undefined to not reconnect.')
}

['bufferedAmount', 'url', 'readyState', 'protocol', 'extensions'].forEach(function(readOnlyProp) {
Object.defineProperty(self, readOnlyProp, {
get: function() { return realWs[readOnlyProp] }
})
})

function clearPendingReconnectIfNeeded() {
if (pendingReconnect) {
clearTimeout(pendingReconnect)
pendingReconnect = null
}
}

var ononline = function(event) {
if (reconnectWhenOnlineAgain) {
clearPendingReconnectIfNeeded()
reconnect(event)
}
},
onoffline = function() {
reconnectWhenOnlineAgain = true
realWs.close(1000)
},
connectivityEventsAttached = false

function detachConnectivityEvents() {
if (connectivityEventsAttached) {
global.removeEventListener('online', ononline)
global.removeEventListener('offline', onoffline)
connectivityEventsAttached = false
}
}

function attachConnectivityEvents() {
if (!connectivityEventsAttached) {
global.addEventListener('online', ononline)
global.addEventListener('offline', onoffline)
connectivityEventsAttached = true
}
}

self.send = function() {
return realWs.send.apply(realWs, arguments)
}

self.close = function(code, reason) {
if (typeof code !== 'number') {
reason = code
code = 1000
}

clearPendingReconnectIfNeeded()
reconnectWhenOnlineAgain = false
explicitlyClosed = true
detachConnectivityEvents()

return realWs.close(code, reason)
}

self.open = function() {
if (realWs.readyState !== WebSocket.OPEN && realWs.readyState !== WebSocket.CONNECTING) {
clearPendingReconnectIfNeeded()
reconnectWhenOnlineAgain = false
explicitlyClosed = false

newWebSocket()
}
}

function reconnect(event) {
if ((!opts.shouldReconnect.handle1000 && event.code === 1000) || explicitlyClosed) {
attempts = 0
return
}
if (navigator.onLine === false) {
reconnectWhenOnlineAgain = true
return
}

var delay = opts.shouldReconnect(event, self)
if (typeof delay === 'number') {
pendingReconnect = setTimeout(newWebSocket, delay)
}
}

Object.defineProperty(self, 'listeners', {
value: {
open: [function(event) {
if (connectTimeout) {
clearTimeout(connectTimeout)
connectTimeout = null
}
event.reconnects = ++reconnects
event.attempts = attempts
attempts = 0
reconnectWhenOnlineAgain = false
}],
close: [reconnect]
}
})

Object.defineProperty(self, 'attempts', {
get: function() { return attempts },
enumerable: true
})

Object.defineProperty(self, 'reconnects', {
get: function() { return reconnects },
enumerable: true
})

function newWebSocket() {
pendingReconnect = null
realWs = new WebSocket(url, protocols || undefined)
realWs.binaryType = self.binaryType

attempts++
self.dispatchEvent(Object.assign(new CustomEvent('connecting'), {
attempts: attempts,
reconnects: reconnects
}))

connectTimeout = setTimeout(function() {
connectTimeout = null
detachConnectivityEvents()
self.dispatchEvent(Object.assign(new CustomEvent('timeout'), {
attempts: attempts,
reconnects: reconnects
}))
}, opts.timeout)

;['open', 'close', 'message', 'error'].forEach(function(stdEvent) {
realWs.addEventListener(stdEvent, function(event) {
self.dispatchEvent(event)

var cb = self['on' + stdEvent]
if (typeof cb === 'function') {
return cb.apply(self, arguments)
}
})
})

if (!opts.ignoreConnectivityEvents) {
attachConnectivityEvents()
}
}

if (opts.automaticOpen) {
newWebSocket()
}
}

RobustWebSocket.defaultOptions = {
// the time to wait before a successful connection
// before the attempt is considered to have timed out
timeout: 4000,
// Given a CloseEvent or OnlineEvent and the RobustWebSocket state,
// should a reconnect be attempted? Return the number of milliseconds to wait
// to reconnect (or null or undefined to not), rather than true or false
shouldReconnect: function(event, ws) {

// AJH:
/*
if (event.code === 1008 || event.code === 1011) return
return [0, 3000, 10000][ws.attempts]
*/

return 1000;
},

// Flag to control whether attachement to navigator online/offline events
// should be disabled.
ignoreConnectivityEvents: false,

// Create and connect the WebSocket when the instance is instantiated.
// Defaults to true to match standard WebSocket behavior
automaticOpen: true
}

// AJH:
//RobustWebSocket.prototype.binaryType = 'blob'
RobustWebSocket.prototype.binaryType = 'arraybuffer'

// Taken from MDN https://developer.mozilla.org/en-US/docs/Web/API/EventTarget
RobustWebSocket.prototype.addEventListener = function(type, callback) {
if (!(type in this.listeners)) {
this.listeners[type] = []
}
this.listeners[type].push(callback)
}

RobustWebSocket.prototype.removeEventListener = function(type, callback) {
if (!(type in this.listeners)) {
return
}
var stack = this.listeners[type]
for (var i = 0, l = stack.length; i < l; i++) {
if (stack[i] === callback) {
stack.splice(i, 1)
return
}
}
}

RobustWebSocket.prototype.dispatchEvent = function(event) {
if (!(event.type in this.listeners)) {
return
}
var stack = this.listeners[event.type]
for (var i = 0, l = stack.length; i < l; i++) {
stack[i].call(this, event)
}
}

return RobustWebSocket
}, typeof window != 'undefined' ? window : (typeof global != 'undefined' ? global : this));


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
var crc32blk = (function()
{
    var table = new Uint32Array(256);

    // Pre-generate crc32 polynomial lookup table
    // http://wiki.osdev.org/CRC32#Building_the_Lookup_Table
    // ... Actually use Alex's because it generates the correct bit order
    //     so no need for the reversal function
    for(var i=256; i--;)
    {
        var tmp = i;

        for(var k=8; k--;)
        {
            tmp = tmp & 1 ? 3988292384 ^ tmp >>> 1 : tmp >>> 1;
        }

        table[i] = tmp;
    }

    // crc32blk
    // Example input        : [97, 98, 99, 100, 101] (Uint8Array)
    // Example output       : 2240272485 (Uint32)
    return function(crc, data )
    {
		crc = (crc ^ -1) >>> 0;
		
        for(var i=0, l=data.length; i<l; i++)
        {
            crc = crc >>> 8 ^ table[ crc & 255 ^ data[i] ];
        }

        return (crc ^ -1) >>> 0; // Apply binary NOT
    };

})();

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
const RKSYXTBL = 
[
 	["INQUIRY_REQ",				0x00,	"reqflags:u8"],
 	["INQUIRY_RSP",				0x40,	"hwversion:u16 fwversion:u16 mode_and_flags:u8 cur_song:u8 cur_part:u8 dummy1:u8 dummy2:u32 svnstr:str64 serstr:str64"],
 	["FWUPDBLOB_REQ",			0x24,   "dat:blob"],
 	["FWUPDBLOB_RSP",			0x64,	"idcode:u32 res:u8"],
 	["FWUPDSTART_REQ",			0x21,	"dat:blob"],
 	["FWUPDSTART_RSP",			0x61,	"res:u8"],
 	["FWUPDFINISH_REQ",			0x22,	"arg1:u32"],
 	["FWUPDFINISH_RSP",			0x62,	"res:u8"],
 	["SETBITRATE_REQ",			0x23,	"bitrate:u32"],
 	["SETBITRATE_RSP",			0x63,	"res:u8"],
 	["FILEWRITESTART_REQ",		0x25,	"totsz:u32 ftype:u8 idx:u8"],
 	["FILEWRITESTART_RSP",		0x65,	"res:u8"],
 	["FILEREADSTART_REQ",		0x26,	"ftype:u8 idx:u8"],
 	["FILEREADSTART_RSP",		0x66,	"totsz:u32 res:u8"],
 	["FILECLOSE_REQ",			0x27,	""],
 	["FILECLOSE_RSP",			0x67,	"res:u8"],
 	["FILEREAD_REQ",			0x28,	"ofs:u32"],
 	["FILEREAD_RSP",			0x68,	"ofs:u32 crc:msgcrc res:u8 dat:blob"],
 	["FILEWRITE_REQ",			0x29,	"ofs:u32 crc:msgcrc dat:blob"],
 	["FILEWRITE_RSP",			0x69,	"ofs:u32 res:u8"],
];

const RKSYXHDR = [0xF0,0x00,0x21,0x23,0x00];	// :-( unfortunately const doesn't really protect the array's contents here...

const RKFILEIOMAXCHUNKSZ	 = 128;


/**
 * encode sysex data
 * @param {Array} srcdat source data to encdoe
 * @return encoded data
 */
function encodeSysex(srcdat)
{
	var d;
	var msbctr = 8;
	var desdat = [];
	
	for (var srcidx=0; srcidx<srcdat.length; srcidx++)
	{
		if (msbctr >= 7)
		{
			desdat[desdat.length] = 0;
			msbctr = 0;
		}
		d = srcdat[srcidx];
		if (d & 0x80)
		{
			desdat[desdat.length - msbctr - 1] |= (1<<msbctr);
		}
		desdat[desdat.length] = d & 0x7f;
		msbctr ++;			
	}	
	
	return desdat;
}

/**
 * decode sysex data
 * @param srcdat{Array} source data
 * @param offset{Number} offset to start reading from
 * @param n{Number} number of bytes to parse
 * @returns {Array}
 */
function decodeSysex(srcdat, offset, n)
{
	var d;
	var msb;
	var desdat = [];
	
	for (var srcidx=0; srcidx<n; srcidx++)
	{
		d = srcdat[offset+srcidx];
		
		if ((srcidx & 7) == 0)
		{
			msb = d;
		}
		else
		{
			if (msb & 0x01)
			{
				d |= 0x80;
			}
			msb >>= 1;
			desdat[desdat.length] = d;
		}
	}
	
	return desdat;
}

/**
 * Retrokits sysex message
 */
class RKSyxMsg
{
	initFields(idx)
	{
		this._idx = idx;
		var msgdef = RKSYXTBL[idx];
		
		this._type = msgdef[0];
		this._cmd = msgdef[1];

		this._fields = [];
		
		if (msgdef[2] != "")
		{
			var str = msgdef[2].split(" ");
			for (var i=0; i<str.length; i++)
			{
				var fstr = str[i].split(":");
				var field = {};
				
				field.namestr = fstr[0];		// field name
				field.typestr = fstr[1];		// field type
				this._fields[i] = field;
				
				this[field.namestr] = (field.typestr == "str64") ? "" : 0;		// field value
			}
		}
	}
	
	/**
	 * create sysex message
	 * @param {string|array} construction argument
	 */
	constructor(arg)
	{
		// arg can be 
		// a) MIDI sysex message
		// b) string containing name (like "INQUIRY_REQ" or "SETPARAM_REQ")
		
		this._idx = -1;
		this._type = "unknown";
		this._cmd = 0;
		this._fields = [];
		this._machine_id = 0;
		
		//
		// argument passed is sysex message?
		//
		if (typeof arg === "object")
		{
			var syxmsg = arg;
			var match = false;
			
			// minimal message = RKSYXHDR <machine_id> <cmd> 0xF7
			if (syxmsg.length >= (RKSYXHDR.length + 3))
			{
				match = true;
				for(var i=0; i<RKSYXHDR.length; i++)
				{
					if (syxmsg[i] != RKSYXHDR[i])
					{
						match = false;
						break;
					}
				}
				
				if (match)
				{
					var l = RKSYXHDR.length;
					var cmd = syxmsg[RKSYXHDR.length+1] & 0x7f;

					for (var i=0; i<RKSYXTBL.length; i++)
					{
						var msgdef = RKSYXTBL[i];
						if (cmd == msgdef[1])
						{
							// init field structure
							this.initFields(i);
							
							this._machine_id = syxmsg[RKSYXHDR.length]; 
							
							// decode sysex arguments
							var args = decodeSysex(syxmsg,RKSYXHDR.length+2,syxmsg.length-RKSYXHDR.length-3);
							
							var p = 0;
							
							for (var i=0; i<this._fields.length; i++)
							{
								var f = this._fields[i];
								var val = 0;
								
								switch (f.typestr)
								{
									case "u8":
										if (p < args.length)
										{
											val = args[p];
										}
										p += 1;
										break;
										
									case "u16":
										if (p <= (args.length-2))
										{
											val = args[p] + (args[p+1]<<8);
										}
										p += 2;
										break;
										
									case "u32":
									case "msgcrc":
										if (p<= (args.length-4))
										{
											// GRMBL GRMBL 
											//val= args[p] + (args[p+1]<<8) + (args[p+2]<<16) + (args[p+3]<<24);  
											val= args[p] + (args[p+1]*256) + (args[p+2]*256*256) + (args[p+3]*256*256*256);
											
											// when this is the message CRC: set the bytes to '0' here, so that computing the CRC later on will read these as '0'
											if (f.typestr == "msgcrc")
											{
												args[p] = args[p+1] = args[p+2] = args[p+3] = 0;
											}
										}
										p+=4;
										break;
										
									case "str64":
										if (p < args.length)
										{
											var c;
											val = "";
											do
											{
												c = args[p++];
												if ((c>=32) && (c<128))
												{
													val += String.fromCharCode(c);
												}
											} while ((c != 0) && (p < args.length));  
										}
										break;
										
									case "blob":
										{
											val = [];
											
											if (p < args.length)
											{
												var m = args.length - p;
												
												for (var j=0; j<m; j++)
												{
													val[j] = args[p+j];
												}
												
												p += m;
											}
											
										}
										break;
										
									default:
										console.error("internal error: unknown field type '"+f.typestr+"' encountered during construct from sysex");
										break;
								}
								
								this.setFieldVal(i,val);
							}
							
							// check msgcrc field:  (which is ALWAYS computed over complete message)
							for (var i=0; i<this._fields.length; i++)
							{
								var f = this._fields[i];
								if (f.typestr == "msgcrc")
								{
									var computed_crc = crc32blk(0,args);
									var received_crc = this.getFieldVal(i);
									var crc_ok = computed_crc == received_crc;
									this["crc_ok"] = crc_ok;
									
									break;
								}	
							}
							
							// command is found: abort the search
							break;
						}
					}
				}
			}
		}
		
		//
		// argument passed is a string?
		//
		if (typeof arg === "string")
		{
			// try to find idx in table:
			for (var i=0; i<RKSYXTBL.length; i++)
			{
				var msgdef = RKSYXTBL[i];
				
				if (arg == msgdef[0])
				{
					// init field structure
					this.initFields(i);

					// init field values passed as arguments to the constructor
					for (var j=0; (j<this._fields.length) && (j<(arguments.length-1)); j++)
					{
						this.setFieldVal(j, arguments[j+1]);
					}
					
					break;
				}
			}
		}
	}
	
	get type()
	{
		return this._type;
	}
	
	get machine_id()
	{
		return this._machine_id;
	}
	
	setFieldVal(idx,val)
	{
		this[this._fields[idx].namestr] = val;
	}
	
	getFieldVal(idx)
	{
		return this[this._fields[idx].namestr];
	}
	
	/**
	 * encode message into sysex array
	 */
	encodeToSysex(machine_id)
	{
		var res = RKSYXHDR.slice();
		
		res.push(machine_id);
		res.push(this._cmd);
		
		var msgcrcargsidx = -1;
		
		var args = [];
		for (var i=0; i<this._fields.length; i++)
		{
			var f = this._fields[i];
			var val = this[f.namestr];
			
			switch (f.typestr)
			{
				case "u8":			// 8 bit integer
					args.push(val);
					break;
					
				case "u16":			// 16 bit integer
					args.push(val&0xff);
					args.push((val>>8)&0xff);
					break;
					
				case "u32":			// 32 bit integer
					args.push(val&0xff);
					args.push((val>>8)&0xff);
					args.push((val>>16)&0xff);
					args.push((val>>24)&0xff);
					break;
					
				case "msgcrc":		// checksum
					msgcrcargsidx = args.length; 
					args.push(0);
					args.push(0);
					args.push(0);
					args.push(0);
					break;
					
				case "str64":		// 64 byte string
					{
						// string should be 64 byte long
						var m = 64;
						
						// but if this is the last argument the string may be encoded with it's actual (possibly shorter) length:
						if (i == this._fields.length-1)
						{
							if (val.length < m)
							{
								m = val.length;
							}
						}
						for (var j=0; j<m; j++)
						{
							args.push((j<val.length) ? val.charCodeAt(j) : 0);
						}
					}
					break;
					
				case "blob":		// binary blob
					for (const v of val)
					{
						args.push(v);
					}
					break;
					
				default:
					args.push(0);
					console.error("internal error: unknown field type '"+f.typestr+"' encountered during encodeToSysex()");
					break;
			}
		}
		
		if (msgcrcargsidx >= 0)
		{
			var computed_crc = crc32blk(0,args);
			args[msgcrcargsidx+0] = computed_crc & 255;
			args[msgcrcargsidx+1] = (computed_crc>>8) & 255;
			args[msgcrcargsidx+2] = (computed_crc>>16) & 255;
			args[msgcrcargsidx+3] = (computed_crc>>24) & 255;
		}
		
		res = res.concat(encodeSysex(args));
		
		res.push(0xF7);
		
		return res;
	}
	
	/**
	 * (try to) decode & construct message from sysex
	 */
	static decodeFromSysex(syxmsg)
	{
		var res = new RKSyxMsg(syxmsg);
		
		return (res.type == "unknown") ? null : res;
	}
	
	toString()
	{
		var str = this.type;
		
		for (var i=0; i<this._fields.length; i++)
		{
			var f = this._fields[i];
			
			if (i==0)
			{
				str += ":";
			}
			
			str += " " + f.namestr + "=";
			
			if (f.typestr != "blob")
			{
				var val = this[f.namestr];
				if (typeof val == "string")
				{
					str += val;
				}
				else
				{
					str += "0x"+this[f.namestr].toString(16);
				}
			}
			else
			{
				str += "<blob sz="+this[f.namestr].length+">";
			}
		}
		
		return str;
	}
}

function encodeVsStr(vs)
{
	return "" + (vs>>8) + "." + ((vs>>4)&0x0f) + (vs&0x0f);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
const RKDEV_PINGTIME_MS	= 1000;

const RKDEV_STATE_IDLE = "IDLE";				// device is not communicating
const RKDEV_STATE_BUSY = "BUSY";				// device is setting up connection
const RKDEV_STATE_OPEN = "OPEN";				// device is fully open for communications
const RKDEV_STATE_FWUP = "FWUP";				// device is currently updating firmware
const RKDEV_STATE_EROR = "EROR";				// device has communication error
const RKDEV_STATE_FILE = "FILE";				// device is busy performing file operation

const RKDEV_SHOW_MIDI = 1;

const RKDEV_BLOB_RETRY	= 10;

const RKDEV_FILE_START_TIMEOUT_MS	= 1000;
const RKDEV_FILE_IO_TIMEOUT_MS		= 3000;
const RKDEV_FILE_RETRY 				= 4;


class RKDev
{
	constructor(machine_id)
	{
		this._observers = [];
		this._io = null;
		this._machine_id = machine_id
		this._fwversion = "";
		this._hwversion = "";
		this._svnstr = "";
		this._serstr = "";
		this._fwversionstr = "";
		this._hwversionstr = "";
		this._state = "";
		this._ping = false;
		this._fwupdtmr = null;
		this._pingtmr = null;
		this._pingctr = 0;
		this._errctr = 0;
		this._file_tmr = null;
		this._file_type = 0;
		this._file_idx = 0;
		this._file_blob = null;
		this._file_state = "";
		this._file_retry = 0;
		this._file_is_writing = false;
		this._file_promise_resolve = null;
		this._file_ofs = 0;
		this._file_sz = 0;
		this._file_progress = 0;
		
		this.pingInit();
	}
	
	addObserver(o)
	{
		this._observers.push(o);
	}
	
	setIO(io)
	{
		this._io = io;
		this._io.setParent(this);
	}
	
	start()
	{
		this.state = RKDEV_STATE_IDLE;
		this._io.start();
	}
	
	reStart()
	{
		this._io.reStart();
	}
	
	sendBtn(idx, v)
	{
		this._io.sendBtn(idx,v);
	}
	
	notifyObservers(functionname, ...args)
	{
		for (const o of this._observers)
		{
			if (typeof o[functionname] === "function")
			{
				o[functionname](...args);
			}
		}
	}
	
	get fwversion()
	{
		return this._fwversion;
	}
	
	get fwversionstr()
	{
		return this._fwversionstr;
	}
	
	get hwversionstr()
	{
		return this._hwversionstr;
	}
	
	get hwserialstr()
	{
		return this._serstr;
	}
	
	onRxMIDISyx(syxmsg)
	{
		if (RKDEV_SHOW_MIDI)
		{
			RKTRACE(RKTRACE_RKDEV,"<-- "+syxmsg);
		}
		
		switch (syxmsg.type)
		{
			case "INQUIRY_RSP" :
				//this.stopQueueingInquiry();
				this._fwversion = syxmsg.fwversion;
				this._hwversion = syxmsg.hwversion;
				this._svnstr = syxmsg.svnstr;
				this._serstr = syxmsg.serstr == "" ? "unknown" : syxmsg.serstr;
				
				this._fwversionstr = encodeVsStr(this._fwversion);
				this._hwversionstr = encodeVsStr(this._hwversion);
				
				this.pingctr = 3;
				this.notifyObservers("onINQUIRY_RSP");
				break;
				
			case "FWUPDBLOB_RSP" :
				this.fwupdOnBlobRsp(syxmsg);
				break;
				
			case "FWUPDSTART_RSP" :
				this.fwupdOnStartRsp(syxmsg);
				break;
				
			case "FWUPDFINISH_RSP" :
				this.fwupdOnFinishRsp(syxmsg);
				break;
				
			case "FILEREADSTART_RSP":
				this.fileOnReadStartRsp(syxmsg);
				break;
				
			case "FILEWRITESTART_RSP":
				this.fileOnWriteStartRsp(syxmsg);
				break;
				
			case "FILECLOSE_RSP":
				this.fileOnCloseRsp(syxmsg);
				break;
				
			case "FILEREAD_RSP":
				this.fileOnReadRsp(syxmsg);
				break;
				
			case "FILEWRITE_RSP":
				this.fileOnWriteRsp(syxmsg);
				break;
				
			default:
				break;
		}
	}
	
	onRxMIDI(msg)
	{
		//this._parent.notifyObservers("onMIDIIOMsg",ev.data);
		
		if (msg[0] == 0xf0)
		{
			var syxmsg = RKSyxMsg.decodeFromSysex(msg);
			
			if (syxmsg != null)
			{
				this.onRxMIDISyx(syxmsg);
			}
		}
	}
	
	txMIDI(req)
	{
		if ('type' in req)
		{
			if (req.type != "unknown")
			{
				if (RKDEV_SHOW_MIDI)
				{
					RKTRACE(RKTRACE_RKDEV,"--> "+req);
				}
				this._io.txMIDI(req.encodeToSysex(this._machine_id));
			}
			else
			{
				console.error("trying to send 'unknown' request");
			}
		}
		else
		{
			this._io.txMIDI(req);
		}
	}
	
	//////////////////////////////////// STATE //////////////////////////////////
	set state(v)
	{
		if (v != this._state)
		{
			this._state = v;
			
			switch (v)
			{
				case RKDEV_STATE_OPEN:
					RKTRACE(RKTRACE_RKDEV,"connected to device: fwversion="+this._fwversionstr+", hwversion="+this._hwversionstr+" svnstr=\""+this._svnstr+"\" serstr=\""+this._serstr+"\"");
					break;
					
				case RKDEV_STATE_BUSY:
					break;
					
				case RKDEV_STATE_EROR:
					RKTRACE(RKTRACE_RKDEV,"connection lost");
					break;
					
				default:
					break;
			}
			
			this.notifyObservers("onDevState",v);
		}
	}
	
	get state()
	{
		return this._state;
	}
	
	get isopen()
	{
		return this._state == RKDEV_STATE_OPEN;
	}
	
	onIOStateChanged(newstate)
	{
		//RKTRACE(RKTRACE_RKDEV,"onIOStateChanged() newstate="+newstate);
		
		switch (newstate)
		{
			case RKIO_STATE_IDLE:
			default:
				this.state = RKDEV_STATE_IDLE;
				break;
				
			case RKIO_STATE_BUSY:
				this.state = RKDEV_STATE_BUSY;
				break;
				
			case RKIO_STATE_OPEN:
				this.state = RKDEV_STATE_BUSY;
				this.pingSendDevInqReq();
				break;
		}
	}
	
	//////////////////////////////////// PING ///////////////////////////////////
	pingInit()
	{
		const self = this;
		setInterval(function(){self.pingOnInterval()}, RKDEV_PINGTIME_MS);
	}
	
	pingSendDevInqReq()
	{
   		let syxmsg = new RKSyxMsg("INQUIRY_REQ");
   		syxmsg.reqflags = 2;
   		this.txMIDI(syxmsg);
	}
	
	pingOnInterval()
	{
		if (!this.fwupdisbusy && !this.fileisbusy)
		{
			if (this._io.isopen)
			{
				this.pingSendDevInqReq();
		   		
		   		if (this.pingctr > 0)
		   		{
		   			this.pingctr--;
		   		}
		   		
		   		if ((this.pingctr == 0) && (this.state == RKDEV_STATE_EROR))
		   		{
					this._io.reStart();
		   		}
			}
		}
	}
	
	set pingctr(v)
	{
		if (v != this._pingctr)
		{
			this._pingctr = v;
			switch (this.state)
			{
				default:
					break;
					
				case RKDEV_STATE_OPEN:
					if (v == 0)
					{
						this.state = RKDEV_STATE_EROR;
					}
					break;
					
					
				case RKDEV_STATE_BUSY:
				case RKDEV_STATE_EROR:
					if (v != 0)
					{
						this.state = RKDEV_STATE_OPEN;
					}
					break;
			}
		}
	}
	
	get pingctr()
	{
		return this._pingctr;
	}

	//////////////////////////////////// FW UPDATE /////////////////////////////
	get fwblobs()
	{
		return this._fwblobs;
	}
	
	get fwblobidx()
	{
		return this._fwblobidx;
	}
	
	set fwblobidx(v)
	{
		this._fwblobidx = v;
	}
	
	get fwnmbparts()
	{
		return this._fwnmbparts;
	}
	
	get fwpartidx()
	{
		return this._fwpartidx;
	}
	
	get fwupdtmr()
	{
		return this._fwupdtmr;
	}
	
	get fwblobretry()
	{
		return this._fwblobretry;
	}
	
	set fwblobretry(v)
	{
		this._fwblobretry = v;
	}
	
	get fwupdisbusy()
	{
		//return this.fwblobidx >= 0;
		return this.fwupdtmr != null;
	}
	
	set fwupdtmr(v)
	{
		// always clear timer
		if (this._fwupdtmr)
		{
			clearTimeout(this._fwupdtmr);
			this._fwupdtmr = null;
		}
		
		if (v)
		{
			const self = this;
			this._fwupdtmr = setTimeout(function(){self.fwupdOnTimeout();},v);
		}
	}
	
	updateFirmwareFromByteArray(v)
	{
		var res = decodeSMF(v);
				
		if (res == 0)
		{
			this._fwblobs = [];
			
			for (const ev of midiFile.tracks[0].events)
			{
				if (ev.type == "sysex")
				{
					var rksyxmsg = new RKSyxMsg(ev.metaData);
					if ((rksyxmsg.type == "FWUPDBLOB_REQ") || (rksyxmsg.type == "FWUPDSTART_REQ") || (rksyxmsg.type == "FWUPDFINISH_REQ"))
					{
						if (rksyxmsg.machine_id == this._machine_id)
						{
							this._fwblobs.push(rksyxmsg);
						}
					}
				}
			}
			
			this.fwupdStart();
		}
		else
		{
			RKTRACE(RKTRACE_RKDEV,"error reading firmware file");
		}
	}


	updateFirmwareFromXHRSMF(fname)
	{
		var xhr = new XMLHttpRequest();
		xhr.open('GET', './'+fname,true);
		xhr.responseType='arraybuffer';
		
		const self = this;
		
		xhr.onreadystatechange = function() 
		{
			if(xhr.readyState === 4 && xhr.status === 200) 
			{
				self.updateFirmwareFromByteArray(this.response);
			}
			else
			{
				RKTRACE(RKTRACE_RKDEV,"updateFirmwareFromXHRSMF() readyState="+xhr.readyState+", status="+xhr.status);
			}
		}
		xhr.send();
	}
	updateFirmware(arg)
	{
		if (!this.fwupdisbusy)
		{
			if (typeof arg == "string")
			{
				RKTRACE(RKTRACE_RKDEV,"updating firmware from '"+arg+"'");
				this.updateFirmwareFromXHRSMF(arg);
			}
			else
			{
				this.updateFirmwareFromByteArray(arg);
			}
		}
	}
	/* SyX PARSING */
	sendSyxFromByteArray(v)
	{
		var res = v;
				// *******
		if(v.byteLength>0){
			// do it
			var syxView = new Uint8Array(v);
			let packetReps=syxView.length;
			this.txMIDI(syxView);
			/*
			const syxTimer = setInterval(() => {
			  if (packetReps==0) {
			    clearInterval(syxTimer);
			    return;
			  }else{
			  	this.txMIDI(syxView[syxView.length-packetReps]);
				//console.log("data:"+ (syxView.length-packetReps)+ " : "+ syxView[syxView.length-packetReps]);
			  	packetReps--;
		  	  }
			}, 5);
			*/
			//RKTRACE(RKTRACE_RKDEV,"len:"+ v.byteLength);
		}
		else
		{
			RKTRACE(RKTRACE_RKDEV,"error reading syx file");
		}
	}
	
	updateSyxFromXHRSMF(fname)
	{
		var xhr = new XMLHttpRequest();
		xhr.open('GET', './'+fname,true);
		xhr.responseType='arraybuffer';
		
		const self = this;
		
		xhr.onreadystatechange = function() 
		{
			if(xhr.readyState === 4 && xhr.status === 200) 
			{
				self.sendSyxFromByteArray(this.response);
			}
			else
			{
				RKTRACE(RKTRACE_RKDEV,"updateSyxFromXHRSMF() readyState="+xhr.readyState+", status="+xhr.status);
			}
		}
		xhr.send();
	}
	/* SyX PARSING */

	parseSyx(arg)
	{
		if (!this.fwupdisbusy)
		{
			if (typeof arg == "string")
			{
				RKTRACE(RKTRACE_RKDEV,"parsing SYX file from '"+arg+"'");
				this.updateSyxFromXHRSMF(arg);
			}
			else
			{
				this.sendSyxFromByteArray(arg);
			}
		}
	}

	fwupdStart()
	{
        RKTRACE(RKTRACE_RKDEV,"starting firmware update with "+this.fwblobs.length+" blobs");
        
        if (this.state == RKDEV_STATE_OPEN)
        {
            this.notifyObservers("onFWUPD_START");
            this.notifyObservers("onFWUPD_STS","starting",1);

            this._fwnmbparts = 0;
            this._fwupdprogress = 0;
            for (const b of this.fwblobs)
            {
                if (b.type == "FWUPDFINISH_REQ")
                {
                    this._fwnmbparts++;
                }
            }
            this._fwpartidx = 0;
            this._fwblobidx = -1;
            this._fwblobretry = 0;
            this._errctr = 0;

            if (this._fwnmbparts != 1)
            {
                this.fwupdFinished("update FAILED: invalid file",0);
            }
            else
            {
                this.state = RKDEV_STATE_FWUP;
                this.fwupdQueueBlobReq(0);
            }
        }
        else
        {
        	this.fwupdFinished("device not opened",0);
        }
	}
	
	fwupdQueueBlobReq(idx)
	{
		if (idx < this.fwblobs.length)
		{           
			if (idx == this.fwblobidx)
			{
				// it's a retry:
				if (this.fwblobretry < RKDEV_BLOB_RETRY)
				{
					this.fwblobretry++;
					RKTRACE(RKTRACE_RKDEV,"RETRY #"+this.fwblobretry+"/"+RKDEV_BLOB_RETRY+" of blob "+idx);
				}
				else
				{
					// out of retries:
					this.fwupdFinished("update FAILED",0);
					return;
				}
			} 
			else
			{
				this.fwblobretry = 0;
			}
			
			this.fwblobidx = idx;
            var blob = this.fwblobs[this.fwblobidx];

            // set time out depending on blob type:
            var timeout_ms = 200;
            if (blob.type == "FWUPDSTART_REQ")
            {
                timeout_ms = 15000;
                this.notifyObservers("onFWUPD_STS","preparing update",1);
            }
            else
            if (blob.type == "FWUPDFINISH_REQ")
            {
            	timeout_ms = 5000;
            }
            else
            {
            	let p = Math.floor(this.fwblobidx * 100 / this.fwblobs.length);
            	if (p == 0)
            	{
            		p = 1;
            	}
            	if (p != this._fwupdprogress)
            	{
            		this._fwupdprogress = p;
                	this.notifyObservers("onFWUPD_STS","transferring data",this._fwupdprogress);
            	}
            }
            this.fwupdtmr = timeout_ms;
			this.txMIDI(this.fwblobs[this.fwblobidx]);
		}
	}
	
	fwupdFinished(msg,progress)
	{
		this.fwupdtmr = null;
		this.fwblobidx = -1;
		
		RKTRACE(RKTRACE_RKDEV,"fwupdFinished() (completed with "+this._errctr+" transfer errors)");
		
		this.notifyObservers("onFWUPD_STS",msg,progress);
        this.notifyObservers("onFWUPD_FINISH");
        
        if (this.state == RKDEV_STATE_FWUP)
        {
        	this.state = RKDEV_STATE_OPEN;
        }
	}
	
	fwupdRetry()
	{
		if ((this.state == RKDEV_STATE_FWUP) && (this.fwblobidx > 0))
		{
			// retry
			this._errctr++;
			this.fwupdQueueBlobReq(this.fwblobidx);
		}
		else
		{
			// fail
			this.fwupdFinished("update FAILED",0);
		}
	}
	
	fwupdOnTimeout()
	{
		// blob timeout failure
		RKTRACE(RKTRACE_RKDEV,"fwupdOnTimeout()");
		this.fwupdRetry();
	}
	
	fwupdOnBlobRsp(rsp)
	{
		if (this.fwupdtmr)
		{
			if (rsp.res == 0)
			{
				// stop fwupdtmr:
				this.fwupdtmr = 0;
				
				// blob transferred OK, queue next one
				this.fwupdQueueBlobReq(this.fwblobidx+1);
			}
			else
			{
				// failure: ay ay.. just ignore the response, and let fwupdOnTimeout handle it
				RKTRACE(RKTRACE_RKDEV,"retrying previous block");
				if (this.fwblobidx>1)
				{
					this.fwblobidx--;
				}
				this.fwupdRetry();
			}
		}
	}
	
	fwupdOnStartRsp(rsp)
	{
		this.fwupdOnBlobRsp(rsp);
	}
	
	fwupdOnFinishRsp(syxmsg)
	{
		let res = syxmsg.res;
		if (res == 0)
		{
			this.fwupdFinished("updated OK",100);
		}
		else
		{
			this.fwupdFinished("update failed",0);
		}
	}
	
	fwupdAbort()
	{
		this.fwupdFinished("user abort",0);
	}
	
	
	//////////////////////////////////// FILE OPS /////////////////////////////
    set filetmr(v)
	{
		// always clear timer
		if (this._file_tmr)
		{
			clearTimeout(this._file_tmr);
			this._file_tmr = null;
		}
		
		if (v)
		{
			const self = this;
			this._file_tmr = setTimeout(function(){self.fileOnTimeout();},v);
		}
	}
	
	get fileisbusy()
	{
		return this.state == RKDEV_STATE_FILE;
	}

	
	fileDone(res)
	{
		this.state = RKDEV_STATE_OPEN;
		this.filetmr = 0;
		
       	this.notifyObservers("onFILEIO_STS","transfer done",100);
		
		if (this._file_is_writing)
		{
			this._file_promise_resolve(res);
		}
		else
		{
			if (res == 0)
			{
				this._file_promise_resolve(this._file_blob);	
			}
			else
			{
				this._file_promise_resolve(null);
			}
		}
	}
	
	fileReqStart()
	{
       	this.notifyObservers("onFILEIO_STS","starting",0);

		if (this._file_retry < RKDEV_FILE_RETRY)
		{
			this._file_retry++;
			this.filetmr = RKDEV_FILE_START_TIMEOUT_MS;

			// transmit the request:
			var syxmsg;
			if (this._file_is_writing)
			{
				syxmsg = new RKSyxMsg("FILEWRITESTART_REQ");
				syxmsg.totsz = this._file_sz;
				this._file_state = "EXPECT_WRITESTART_RSP";
			}			
			else
			{
				syxmsg = new RKSyxMsg("FILEREADSTART_REQ");
				this._file_state = "EXPECT_READSTART_RSP";
			}
			
	  		syxmsg.ftype = this._file_type;
	  		syxmsg.idx = this._file_idx;
		 	this.txMIDI(syxmsg);
		}
		else
		{
			this.fileDone(-1);
		}
	}
	
	fileReqClose()
	{
		if (this._file_retry < RKDEV_FILE_RETRY)
		{
			this._file_retry++;
			this.filetmr = RKDEV_FILE_START_TIMEOUT_MS;
			this._file_state = "EXPECT_CLOSE_RSP";
			
			let syxmsg = new RKSyxMsg("FILECLOSE_REQ");
			this.txMIDI(syxmsg);
			
		}
		else
		{
			this.fileDone(-1);
		}
	}
	
	fileReqFileIO()
	{
		if (this.state != RKDEV_STATE_FILE)
		{
			return;
		}		
		
		if (this._file_retry < RKDEV_FILE_RETRY)
		{
			this._file_retry++;
			this._file_state = "EXPECT_FILE_IO_RSP";
			this.filetmr = RKDEV_FILE_IO_TIMEOUT_MS;
						
			if (this._file_is_writing)
			{
		  		let syxmsg = new RKSyxMsg("FILEWRITE_REQ");
		  		syxmsg.ofs = this._file_ofs;
		  		var m = this._file_blob.byteLength - this._file_ofs;
		  		if (m > RKFILEIOMAXCHUNKSZ)
		  		{
					m = RKFILEIOMAXCHUNKSZ;
				}
				syxmsg.dat = new Uint8Array(m);
				for (var i=0; i<m; i++)
				{
					syxmsg.dat[i] = this._file_blob[this._file_ofs + i];
				}
				this.txMIDI(syxmsg);
			}
			else
			{
		  		let syxmsg = new RKSyxMsg("FILEREAD_REQ");
		  		syxmsg.ofs = this._file_blob.length;
			 	this.txMIDI(syxmsg);
			}
			
			if (this._file_sz > 0)
			{
	        	let p = Math.floor(this._file_ofs * 100 / this._file_sz);
	        	if (p == 0)
	        	{
	        		p = 1;
	        	}
	        	if (p != this._file_progress)
	        	{
					this._file_progress = p;
	            	this.notifyObservers("onFILEIO_STS","transferring data",this._file_progress);
				}
	        }
		}
		else
		{
			this.fileDone(-1);
		}
	}
	
	fileRead(filetype, fileidx)
	{
		const self = this;
			
		var promise = new Promise(function(resolve, reject)
		{
			self._file_promise_resolve = resolve;
			self._file_promise_reject = reject;	
		});
		
        if (this.state == RKDEV_STATE_OPEN)
        {
			this.state = RKDEV_STATE_FILE;
			this._file_blob = new Uint8Array();
			this._file_type = filetype;
			this._file_idx = fileidx;
			this._file_retry = 0;
			this._file_is_writing = false;
			this._file_ofs = 0;
			this._file_sz = 0;
			this._file_progress = 1;
			this.fileReqStart();
		}
		else
		{
			self._file_promise_reject();
		}
		
		return promise;
	}
	
	fileWrite(filetype, fileidx, dat)
	{
		const self = this;
		
		var promise = new Promise(function(resolve, reject)
		{
			self._file_promise_resolve = resolve;
			self._file_promise_reject = reject;	
		});
		
        if (this.state == RKDEV_STATE_OPEN)
        {
			this.state = RKDEV_STATE_FILE;
			this._file_blob = dat;
			this._file_type = filetype;
			this._file_idx = fileidx;
			this._file_retry = 0;
			this._file_is_writing = true;
			this._file_ofs = 0;
			this._file_sz = this._file_blob.byteLength;
			this._file_progress = 1;
			this.fileReqStart();
		}
		else
		{
			self._file_promise_reject();
		}
		
		return promise;
	}
	
	fileOnTimeout()
	{
		if (this.state != RKDEV_STATE_FILE)
		{
			return;
		}
		
		switch (this._file_state)
		{
			case "EXPECT_OPEN_RSP":
				this.fileReqStart();
				break;
				
			case "EXPECT_FILE_IO_RSP":
				this.fileReqFileIO();
				break;
				
			default:
				break;
		}
	}
	
	fileOnReadStartRsp(syxmsg)
	{
		if ((this.state != RKDEV_STATE_FILE) || (this._file_state != "EXPECT_READSTART_RSP"))
		{
			return;
		}

		let res = syxmsg.res;
		
		if (res != 0)		
		{
			// try again:
			this.fileReqStart();
		}
		else
		{
			// file opened successfully -> start transfer data
			this._file_sz = syxmsg.totsz;
			this._file_retry = 0;
			this.fileReqFileIO();
			
			if (this._file_sz == 0)
			{
				this._file_progress = null;
	           	this.notifyObservers("onFILEIO_STS","transferring data",this._file_progress);
			}
			
		}
	}
	
	fileOnWriteStartRsp(syxmsg)
	{
		if ((this.state != RKDEV_STATE_FILE) || (this._file_state != "EXPECT_WRITESTART_RSP"))
		{
			return;
		}

		let res = syxmsg.res;
		
		if (res != 0)		
		{
			// try again:
			this.fileReqStart();
		}
		else
		{
			// file opened successfully -> start transfer data
			this._file_retry = 0;
			this.fileReqFileIO();
		}
	}
	
	fileOnReadRsp(syxmsg)
	{
		if ((this.state != RKDEV_STATE_FILE) || (this._file_state != "EXPECT_FILE_IO_RSP") || (this._file_is_writing))
		{
			return;
		}
		
		if (syxmsg.res != 0)
		{
			// retry:
			this.fileReqFileIO();			
		}
		else
		{
			// is this EOF?
			if (syxmsg.dat.length == 0) 
			{
				// we're done with the data: close file
				this._file_retry = 0;
				this.fileReqClose();
			}
			else
			{
				if (syxmsg.crc_ok)
				{
					// add to received blob:
					let n = syxmsg.dat.length;
					let arr = new Uint8Array(this._file_blob.length + n);
					arr.set(this._file_blob);
					for (var i=0; i<n; i++)
					{
						arr[this._file_blob.length+i] = syxmsg.dat[i];
					}	
					this._file_blob = arr;
					this._file_ofs += n;
					
					// continue:
					this._file_retry = 0;
					this.fileReqFileIO();
				}
				else
				{
					// retry 
					this.fileReqFileIO();
				}				
			}
			
		}	
	}
	
	fileOnWriteRsp(syxmsg)
	{
		if ((this.state != RKDEV_STATE_FILE) || (this._file_state != "EXPECT_FILE_IO_RSP") || (!this._file_is_writing))
		{
			return;
		}
		
		if (syxmsg.res != 0)
		{
			// retry:
			this.fileReqFileIO();			
		}
		else
		{
			if (syxmsg.ofs != this._file_ofs)
			{
				// retry
				this.fileReqFileIO();			
			}
			else
			{
		  		var m = this._file_blob.byteLength - this._file_ofs;
		  		if (m > RKFILEIOMAXCHUNKSZ)
		  		{
					m = RKFILEIOMAXCHUNKSZ;
				}
				this._file_ofs += m;
				
				if (this._file_ofs >= this._file_blob.byteLength)
				{
					// we're done with the data: close file
					this._file_retry = 0;
					this.fileReqClose();
				}
				else
				{
					// continue: 
					this._file_retry = 0;
					this.fileReqFileIO();
				}
			}
		}
	}
	
	fileOnCloseRsp(syxmsg)
	{
		if ((this.state != RKDEV_STATE_FILE) || (this._file_state != "EXPECT_CLOSE_RSP"))
		{
			return;
		}
		
		if (syxmsg.res == 0)
		{
			this.fileDone(0);
		}
	}

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// RK008
//
const RK008_MACHINE_ID = 0x08;


const RK008_FILETYPE_GLOBAL 	= 1;
const RK008_FILETYPE_PROJECT 	= 2;
const RK008_FILETYPE_SMF		= 3;

class RK008 extends RKDev
{
	constructor()
	{
		super(RK008_MACHINE_ID);
		this._cur_song = -1;
		this._cur_part = -1;
		this._mode_and_flags = -1;
	}
	
	async getSong(idx)
	{
		if (idx < 0)
		{
			idx = 255;
		}
		const res = await this.fileRead(RK008_FILETYPE_PROJECT,idx);
		
		return res;
	}
	
	async putSong(idx, data)
	{
		if (idx < 0)
		{
			idx = 255;
		}
		const res = await this.fileWrite(RK008_FILETYPE_PROJECT,idx,data);
		
		return res;
	}
	
	async putSMF(idx,data)
	{
		if (idx < 0)
		{
			idx = 255;
		}
		const res = await this.fileWrite(RK008_FILETYPE_SMF,idx,data);
		return res;
	}
	
	async getSMF(idx)
	{
		if (idx < 0)
		{
			idx = 255;
		}
		
		var res = await this.fileRead(RK008_FILETYPE_SMF,idx);
		
		if ((res != null) && (res.length > 0))
		{
			const n_tracks = res[res.length-1];
			
			const tablebeg = res.length - ((n_tracks*4)+4+1);
			
			RKTRACE(RKTRACE_RK008,"SMF contains "+n_tracks+" tracks");
			
			// patch SMF track chunks lengths:
			if (tablebeg >= 14)
			{
				const MTln = "MTln";
				
				if ((res[tablebeg+0] == MTln.charCodeAt(0)) && (res[tablebeg+1] == MTln.charCodeAt(1)) && (res[tablebeg+2] == MTln.charCodeAt(2)) && (res[tablebeg+3] == MTln.charCodeAt(3)) )
				{
					var p = 14;
					
					for (var i=0; i<n_tracks; i++)
					{
						var l = (res[tablebeg+4+(i*4)+0]<<24) + (res[tablebeg+4+(i*4)+1]<<16) + (res[tablebeg+4+(i*4)+2]<<8) + (res[tablebeg+4+(i*4)+3]<<0);
						RKTRACE(RKTRACE_RK008,"  track["+i+"].size="+l+" bytes");
						
						const MTrk = "MTrk";
						
						if ((res[p+0] == MTrk.charCodeAt(0)) && (res[p+1] == MTrk.charCodeAt(1)) && (res[p+2] == MTrk.charCodeAt(2)) && (res[p+3] == MTrk.charCodeAt(3)))
						{
							res[p + 4] = (l>>24) & 255;
							res[p + 5] = (l>>16) & 255;
							res[p + 6] = (l>>8) & 255;
							res[p + 7] = (l>>0) & 255;
							
							p += l + 8;  
						}
						else
						{
							res = null;
							break;
						}
					}
					
					// slice-away length table
					if (res != null)
					{
						res = res.slice(0,tablebeg);
					}
				}
				else
				{
					res = null;
				} 									
			}
			else
			{
				res = null;
			}
		}
		else
		{
			res = null;
		}

		if (res == null)
		{
			RKTRACE(RKTRACE_RK008,"failed to patch SMF length data");
		}
		
		return res;
	}
	
	onRxMIDISyx(syxmsg)
	{
		switch (syxmsg.type)
		{
			case "INQUIRY_RSP" :
				if (syxmsg.mode_and_flags == 0xff)
				{
					this._mode_and_flags = -1;
					this._cur_song = -1;
					this._cur_part = -1;
				}
				else
				{
					this._mode_and_flags = syxmsg.mode_and_flags;
					this._cur_song = syxmsg.cur_song;
					this._cur_part = syxmsg.cur_part;					
				}
				break;
		}
		
		super.onRxMIDISyx(syxmsg);
	}

	get cur_mode()
	{
		// 0 = PART
		// 1 = SONG
		// 2 = STEP
		return (this._mode_and_flags >= 0) ? (this._mode_and_flags & 0x03) : 0;
	}
	
	get is_blocked_for_file_io()
	{
		return (this._mode_and_flags & (1<<2)) != 0;		
	}
	
	get cur_song()
	{
		return this._cur_song;
	}
	
	get cur_part()
	{
		return this._cur_part;		
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// generic I/O class
//
const RKIO_STATE_IDLE	= "IDLE"; 		// stopped
const RKIO_STATE_BUSY	= "BUSY";		// busy setting up IO connection
const RKIO_STATE_OPEN	= "OPEN";		// IO connection is opened

class RKIO
{
	constructor()
	{
		this._parent = null;
		this._state = 0;
	}
	
	setParent(parent)
	{
		this._parent = parent;
	}
	
	set state(v)
	{
		if (v != this._state)
		{
			this._state = v;
			this._parent.onIOStateChanged(v);
		}
	}
	
	get state()
	{
		return this._state;
	}
	
	get isopen()
	{
		return this.state == RKIO_STATE_OPEN;
	}
	
	get isbusy()
	{
		return this.state == RKIO_STATE_BUSY;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// WebSocket I/O
//
const WSPCK_TXT		= 1;
const WSPCK_IND		= 2;
const WSPCK_BTN		= 3;
const WSPCK_TVU		= 4;

class RKIOWebSocket extends RKIO
{
	constructor()
	{
		super();
		this.socket = null;
		this.host = location.host;
	}
	
	start()
	{
		if (this.socket)
		{
			this.socket = null;
		}
		
		this.socket = new RobustWebSocket("ws://"+this.host+"/ws");
		
		const self = this;
		const parent = this._parent;
		
		this.socket.addEventListener("open", function(event)
			{
				RKTRACE(RKTRACE_RK008,"socket open");
				self.state = RKIO_STATE_OPEN;
			});
		
		this.socket.addEventListener("message",function(ev)
			{
				//RKTRACE(RKTRACE_RK008,"socket message: "+event.data);
				
				// binary data? 
				if (ev.data instanceof ArrayBuffer)
				{
					let msg = new Uint8Array(ev.data);
					let msgidx = 0;
					
					//RKTRACE(RKTRACE_RK008,"binary data sz=" + msg.length + " bytes");

					while (msgidx < msg.length)
					{
						let cmd = msg[msgidx++];
						
						switch (cmd)
						{
							case WSPCK_TXT:
								{
									let txt = "";
									let c;
									let idx = msg[msgidx++];
									
									while ((msg.length > msgidx) && ((c = msg[msgidx++]) != 0))
									{
										txt += String.fromCharCode(c);
									}
									parent.notifyObservers("onIO_TXT",idx,txt);
								}
								break;
								
							case WSPCK_IND:
								{
									let idx = msg[msgidx++];
									let val = msg[msgidx++];
									parent.notifyObservers("onIO_IND",idx,val);
								}
								break;
								
							case WSPCK_TVU:
								{
									let msk = msg[msgidx++];
									parent.notifyObservers("onIO_TVU",msk);
								}
								break;
								
								
							default:
								break;
						}
					}
					
					//self.notifyObservers("onSock_midimsg",msg);
					//RKTRACE(RKTRACE_RK008,"RX: midimsg (sz="+msg.length+")");
				}
				else
				{
					// (try to) decode as JSON
					try
					{
						let jsonmsg = JSON.parse(ev.data);
						
						// parse command & arg:
						if (('cmd' in jsonmsg) && ('arg' in jsonmsg))
						{
							parent.notifyObservers("onSock_"+jsonmsg.cmd,jsonmsg.arg);
						}
						
					} catch (err)
					{
					}
				}
			});
		
		this.socket.addEventListener("close", function(event)
			{
				RKTRACE(RKTRACE_RK008,"socket close");
				self.state = RKIO_STATE_IDLE;
			});
		
		this.socket.addEventListener("error", function(event)
			{
				RKTRACE(RKTRACE_RK008,"socket error");
				self.state = RKIO_STATE_IDLE;
			});
		
		parent.addObserver({
			
			onSock_ping:function(arg)
			{
				RKTRACE(RKTRACE_RK008,`RX ping resp=${arg}`);
			},
			
		});
		
		setInterval(function(){self.sockSendMsg("ping");},3000);
	}
	
	sockSendMsg(msg)
	{
		if (this.isopen)
		{
			this.socket.send(msg);
		}
	}
	
	sendBtn(idx, d)
	{
		this.sockSendMsg("btn "+idx+" "+d);
	}
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// IO through MIDI
//
class RKIOMIDI extends RKIO
{
	sendBtn(idx, v)
	{
	}
	
	onRxMIDISyx(syxmsg)
	{
		this._parent.onRxMIDISyx(syxmsg);
	}
	
	onRxMIDI(msg)
	{
		this._parent.onRxMIDI(msg);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// WebMIDI MIDI I/O
//
class RKIOWebMIDI extends RKIOMIDI
{
	constructor()
	{
		super();
		
		this._midiaccess = null;
		this._soll_input_name = null;
		this._soll_output_name = null;
		this._input = null;
		this._output = null;
		this._inputnames = null;
		this._outputnames = null;
		this._observerrefreshtmr = null;
		this._trackstate = {};
		this._trackstate.state = RKIO_STATE_IDLE;
	}
	
	refreshObservers()
	{
		// inputs
		let inputnames_json = JSON.stringify(this._inputnames);
		if (inputnames_json !== this._trackstate.inputnames_json)
		{
			this._trackstate.inputnames_json = inputnames_json;
			this._parent.notifyObservers("onIO_MIDIInputsUpdate");
		}
		
		// outputs
		var outputnames_json = JSON.stringify(this._outputnames);
		if (outputnames_json !== this._trackstate.outputnames_json)
		{
			this._trackstate.outputnames_json = outputnames_json;
			this._parent.notifyObservers("onIO_MIDIOutputsUpdate");
		}
	}
	
	triggerObserverRefresh()
	{
		if (this._observerrefreshtmr != null)
		{
			clearTimeout(this._observerrefreshtmr);
			this._observerrefreshtmr = null;
		}
		
		const self = this;
		this._observerrefreshtmr = setTimeout(function()
				{
					self.refreshObservers();
                },200);
		
		// update state
		this.state = ((this._input !=0) && (this._output != 0)) ? RKIO_STATE_OPEN : RKIO_STATE_IDLE;
	}

	refreshSollInput()
	{
		let selected_input = null;
		
		if ((this._midiaccess != null) && (this._soll_input_name != null))
		{
			for (const it of this._midiaccess.inputs)
			{
				let inp = it[1];
				
				if (inp.name == this._soll_input_name)
				{
					selected_input = inp;
					break;
				}
			}
		}
		
		if (selected_input != this._input)
		{
			// possibly unsusbscribe from old port:
			if (this._input != null)
			{
				this._input.onmidimessage = null;
			}
			
			// from now on: this is the input
			this._input = selected_input;
			
			// subscribe to midi messages coming from port:
			if (this._input != null)
			{
				const self = this;
				this._input.onmidimessage = function(ev)
				{
					self.onMidiMessage(ev);
				}
			}
			
			this.triggerObserverRefresh();
		}
	}
	
	refreshSollOutput()
	{
		let selected_output = null;
		
		if ((this._midiaccess != null) && (this._soll_output_name != null))
		{
			for (const it of this._midiaccess.outputs)
			{
				let out = it[1];
				
				if (out.name == this._soll_output_name)
				{
					selected_output = out;
					break;
				}
			}
		}
		
		if (selected_output != this._output)
		{
			// from now on this is the new output:
			this._output = selected_output;
			this.triggerObserverRefresh();
		}
	}
	
	/**
	 * handle MIDI reception
	 */
	onMidiMessage(ev)
	{
		//console.log("onMidiMessage()");
		if (ev instanceof MIDIMessageEvent)
		{
			if (ev.data[0] != 0xf8)
			{
				RKTRACE(RKTRACE_WEBMIDI,"<-- "+ev.data);
			}
			this._parent.onRxMIDI(ev.data);
		}
	}
	
	/**
	 * refresh inputs
	 */
	refreshInputs()
	{
		this._inputnames = new Array();
		
		if (this._midiaccess != null)
		{
			// note: midiaccess.inputs is a map !
			//       so iterating over it will result in a {key,val} pair!
			for (const it of this._midiaccess.inputs)
			{
				var inp = it[1];
				
				this._inputnames.push(inp.name);
			}
		}
		
		this.refreshSollInput();
		this.triggerObserverRefresh();
	}
	
	/**
	 * refresh outputs
	 */
	refreshOutputs()
	{
		this._outputnames = new Array();
		
		if (this._midiaccess != null)
		{
			// note: midiaccess.outputs is a map !
			//       so iterating over it will result in a {key,val} pair!
			for (const it of this._midiaccess.outputs)
			{
				var out = it[1];
				this._outputnames.push(out.name);
			}
		}
		
		this.refreshSollOutput();
		
		this.triggerObserverRefresh();
	}
	
	/**
	 * called when midi access obtained successfully.
	 * @param midiaccess{MIDIAccess} midiaccess MIDIAccess object obtained
	 */
	onMIDISuccess(midiaccess)
	{
		RKTRACE(RKTRACE_WEBMIDI,"onMIDISuccess");
		this._midiaccess = midiaccess;
		const self = this;
		this._midiaccess.onstatechange = function(event)
		{
			self.onMIDIStateChanged(event);
		}
		this.refreshInputs();
		this.refreshOutputs();
	}
	
	/**
	 * called when failed to obtain MIDI access.
	 */
	onMIDIFailure()
	{
		RKTRACE(RKTRACE_WEBMIDI,"onMIDIFailure");

		this._parent.notifyObservers("onIO_MIDIIOFailure");
	}
	
	onMIDIStateChanged(ev)
	{
		if (ev instanceof MIDIConnectionEvent)
		{
			let port = ev.port;
			if (port.type == "input")
			{
                RKTRACE(RKTRACE_WEBMIDI,"onMIDIStateChanged() : input changed");
				this.refreshInputs();
			}
			if (port.type == "output")
			{
                RKTRACE(RKTRACE_WEBMIDI,"onMIDIStateChanged() : output changed");
				this.refreshOutputs();
			}
        }
        else
        {
            RKTRACE(RKTRACE_WEBMIDI,"onMIDIStateChanged() : unknown event");
        }
	}
	
	/**
	 * @override
	 */
	get inputnames()
	{
		return this._inputnames;
	}
	
	/**
	 * @override
	 */
	selectInput(name)
	{
		this._soll_input_name = name;
		this.refreshSollInput();
	}
	
	/**
	 * @override
	 */
	get outputnames()
	{
		return this._outputnames;
	}
	
	/**
	 * @override
	 */
	selectOutput(name)
	{
		this._soll_output_name = name;
		this.refreshSollOutput();
	}
	
	/**
	 * @override
	 */
	txMIDI(msg)
	{
		if (this._output)
		{
			RKTRACE(RKTRACE_WEBMIDI,"--> "+msg);
			this._output.send(msg);
		}
		else
		{
			RKTRACE(RKTRACE_WEBMIDI,"skipped sending message while no output port active");
		}
	}
	
	start()
	{
		RKTRACE(RKTRACE_WEBMIDI,"starting webmidi I/O");
		
		const self = this;
		
		if (navigator.requestMIDIAccess)
		{
			navigator.requestMIDIAccess(
					{sysex:true}).then
					(
							function(midiaccess)
							{
								self.onMIDISuccess(midiaccess);
							},
							function()
							{
								self.onMIDIFailure();
							}
					);
		}
		else
		{
			this.onMIDIFailure();
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// WebSerial MIDI I/O
//
const WEBSER_DEFAULT_BITRATE	= 38400;
const WEBSER_HIGH_BITRATE		= 230400;
//const WEBSER_HIGH_BITRATE		= 460800;
//const WEBSER_HIGH_BITRATE		= 500000;

class RKIOWebSerial extends RKIOMIDI
{
	constructor()
	{
		super();
		this._syxidx = -1;
		this._syxmsg = [];
		this._writer = null;
		this._reader = null;
	}
	
	async txMIDI(msg)
	{
		if (this.isopen)
		{
			await this._writer.write(Uint8Array.from(msg));
		}
	}
	
	async closeComms()
	{
		this._reader.cancel();
		await this._readuntilclosed;	// <-- note: this will update the iostate !
   		RKTRACE(RKTRACE_WEBSERIAL,"serial port closed");
	}
	
	async readUntilClosed()
	{
   		// Listen to data coming from the serial device.
		try 
		{
	   		while (true) 
	   		{
	   		  const { value, done } = await this._reader.read();
	   		  if (done) 
	   		  {
	   			  // reader has been cancelled
	   			  break;
	   		  }
	   		  
	   		  for (var i=0; i<value.length; i++)
	   		  {
	   			  let d = value[i];
	   			  //RKTRACE(RKTRACE_WEBSERIAL,"RX:"+d);
	   			  
	   			  if (d<0x80)
	   			  {
	   				  // databyte
	   				  if (this._syxidx >= 0)
	   				  {
	   					  this._syxmsg[this._syxidx++] = d;
	   				  }
	   			  }
	   			  else
	   			  {
	   				  // statusbyte
	   				  if (d < 0xf8)
	   				  {
	   					  if (d == 0xf0)
	   					  {
	   						this._syxidx = 0;
   							this._syxmsg = [];
	   						this._syxmsg[this._syxidx++] = 0xf0;
	   					  }
	   					  else
	   					  {
	   						if (this._syxidx > 0)
	   						{
	   							this._syxmsg[this._syxidx++] = 0xf7;
	   							//RKTRACE(RKTRACE_WEBSERIAL,"RX: "+this._syxmsg);
	   							
	   							let syxmsg = RKSyxMsg.decodeFromSysex(this._syxmsg);
	   							
	   							if (syxmsg != null)
	   							{
	   								this.onRxMIDISyx(syxmsg);
	   							}
	   						}
	   						this._syxidx = -1;
	   					  }
	   				  }
	   			  }
	   		   }
	   		}
		}
		catch (error)
		{
			RKTRACE(RKTRACE_WEBSERIAL,"catched reader error" + error);
		}
		finally 
		{
			this.state = RKIO_STATE_BUSY;
			await this._writer.releaseLock();
			await this._reader.releaseLock();
			RKTRACE(RKTRACE_WEBSERIAL,"writer & reader released lock");
		}
		
		await this._port.close();
		RKTRACE(RKTRACE_WEBSERIAL,"port closed");
	}
	
	async startComms(port)
	{
		RKTRACE(RKTRACE_WEBSERIAL,"startComms()");
		
		// already opened? -> close now
		if (this.isopen)
		{
			await this.closeComms();
		}
		
		this.state = RKIO_STATE_BUSY;
		this._port = port;
   		
   		RKTRACE(RKTRACE_WEBSERIAL,"opening serial port @ "+WEBSER_DEFAULT_BITRATE+"bps");
   		await this._port.open({baudRate:WEBSER_DEFAULT_BITRATE, bufferSize:4096});
   		
   		RKTRACE(RKTRACE_WEBSERIAL,"RESETTING TARGET (!) ");
   		await this._port.setSignals({ requestToSend: true, dataTerminalReady: false });
   		await new Promise(resolve => setTimeout(resolve,500));
   		await this._port.setSignals({ requestToSend: false, dataTerminalReady: false });
   		RKTRACE(RKTRACE_WEBSERIAL,"RESET DONE");
   		
   		await new Promise(resolve => setTimeout(resolve,1500));
   		RKTRACE(RKTRACE_WEBSERIAL,"switching bitrate");
   		
   		let syxmsg = new RKSyxMsg("SETBITRATE_REQ");
   		syxmsg.bitrate = WEBSER_HIGH_BITRATE;
   		
   		this._writer = this._port.writable.getWriter();
   		syxmsg = Uint8Array.from(syxmsg.encodeToSysex(this._parent._machine_id));
		await this._writer.write(syxmsg);
		await this._writer.write(syxmsg);
		
   		await new Promise(resolve => setTimeout(resolve,700));
		
		await this._writer.releaseLock();
		
		RKTRACE(RKTRACE_WEBSERIAL,"closing port");

		await this._port.close();
   		
   		RKTRACE(RKTRACE_WEBSERIAL,"opening serial port @ "+WEBSER_HIGH_BITRATE+"bps");
   		await this._port.open({baudRate:WEBSER_HIGH_BITRATE});

   		this._writer = this._port.writable.getWriter();
   		this._reader = this._port.readable.getReader();
   		this._readuntilclosed = this.readUntilClosed();
   		
   		await this._writer.write(syxmsg);
		
   		this.state = RKIO_STATE_OPEN;
	}

	/*  - - - - - - - */	
	/* G changed port user initialization */
	async setupPort()
	{
		if (this.isopen)
		{
			this.closeComms(); //restart init proc
		}
		const filters = [{usbVendorId:0x10c4, usbProductId: 0xea60}]; // filter on CP210X
   	    const port = navigator.serial.requestPort({ filters: filters}).then((port) => 
   	    {
            // Connect to port
			this.startComms(port);
        }).catch((e) => 
        {
		   RKTRACE(RKTRACE_WEBSERIAL,""+e);
           // The user didn't select a port.
		   // dispatch user cancel event
		   this.state = RKIO_STATE_IDLE;
        });
	}
	/*  - - - - - - - */
	
	async reStart()
	{
		if (this.isopen)
		{
			this.startComms(this._port);
		}
	}
	
	start()
	{
	    this.state = RKIO_STATE_BUSY;
		
		navigator.serial.addEventListener('connect', (e) => {
			RKTRACE(RKTRACE_WEBSERIAL,"connect event");
			this.startComms(e.target);
		})
		
		navigator.serial.addEventListener('disconnect', (e) => {
			RKTRACE(RKTRACE_WEBSERIAL,"disconnect event");
			this.state = RKIO_STATE_IDLE;
		})
		
		navigator.serial.getPorts().then((ports) => {
			if (ports.length != 0)
			{
				this.startComms(ports[0]);
			}
			else
			{
				RKTRACE(RKTRACE_WEBSERIAL,"no serial port configured");
			}
		})
	}
}